@@ -18,10 +18,26 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #---------------------------------------------------------------------------
+from threading import Thread
 from base_module import ranaModule
 import geo
 import math
 import time
+
+REROUTE_CHECK_INTERVAL = 5000 # in ms
+#in m/s, about 46 km/h - if this speed is reached, the rerouting threshold is multiplied
+# by REROUTING_THRESHOLD_MULTIPLIER
+INCREASE_REROUTING_THRESHOLD_SPEED = 20
+REROUTING_DEFAULT_THRESHOLD = 30
+# not enabled at the moment - needs more field testing
+REROUTING_THRESHOLD_MULTIPLIER = 1.0
+# how many times needs the threshold be crossed to
+# trigger rerouting
+REROUTING_TRIGGER_COUNT = 3
+
+MAX_CONSECUTIVE_AUTOMATIC_REROUTES = 3
+AUTOMATIC_REROUTE_COUNTER_EXPIRATION_TIME = 600 # in seconds
+
 # only import GKT libs if GTK GUI is used
 from core import gs
 if gs.GUIString == "GTK":
@@ -29,7 +45,7 @@ if gs.GUIString == "GTK":
   import pangocairo
 
 def getModule(m,d,i):
-  return(turnByTurn(m,d,i))
+  return turnByTurn(m,d,i)
 
 class turnByTurn(ranaModule):
   """A turn by turn navigation module."""
@@ -40,11 +56,17 @@ class turnByTurn(ranaModule):
     # initial colors
     self.navigationBoxBackground = (0,0,1,0.3) # very transparent blue
     self.navigationBoxText = (1,1,1,1) # non-transparent white
-
+    self.TBTWorker = None
+    self.TBTWorkerEnabled = False
     self.goToInitialState()
+    self.automaticRerouteCounter = 0 # counts consecutive automatic reroutes
+    self.lastAutomaticRerouteTimestamp = time.time()
+    # reroute even though the route was not yet reached (for special cases)
+    self.overrideRouteReached = False
 
   def goToInitialState(self):
-    self.steps = []
+    """restore initial state"""
+    self.route = None
     self.currentStepIndex = 0
     self.currentStepIndicator = None
     self.espeakFirstAndHalfTrigger = False
@@ -55,7 +77,11 @@ class turnByTurn(ranaModule):
     self.navigationBoxHidden = False
     self.mRouteLength = 0
     self.locationWatchID = None
-
+    self.onRoute = False
+    #rerouting is enabled once the route is reached for the first time
+    self.routeReached = False
+    self.reroutingThresholdMultiplier = 1.0
+    self.reroutingThresholdCrossedCounter = 0
 
   def firstTime(self):
     icons = self.m.get('icons', None)
@@ -73,19 +99,12 @@ class turnByTurn(ranaModule):
         self.startTBT(fromWhere)
     elif message == 'stop':
       self.stopTBT()
-    elif message == 'reroute':
-      # 1. say rerouting is in progress
-      voiceMessage = "rerouting"
-      voice = self.m.get('voice', None)
-      if voice:
-        voice.say(voiceMessage, "en") # make sure rerouting said with english voice
-      time.sleep(2) #TODO: improve this
-      # 2. get a new route from current position to destination
-      self.sendMessage("ms:route:reroute:fromPosToDest")
-      # 3. restart routing for to this new route from the closest point
-      self.sendMessage("ms:turnByTurn:start:closest")
+    elif message == 'reroute': # manual rerouting
+      # reset automatic reroute counter
+      self.automaticRerouteCounter = 0
+      self._reroute()
     elif message == "toggleBoxHiding":
-      print "turnByTurn: toggling navigation box visibility"
+      print("turnByTurn: toggling navigation box visibility")
       self.navigationBoxHidden = not self.navigationBoxHidden
     elif message == "switchToPreviousTurn":
       self.switchToPreviousStep()
@@ -94,20 +113,67 @@ class turnByTurn(ranaModule):
     elif message == "showMessageInsideNotification":
       currentStep = self.getCurrentStep()
       if currentStep:
-        message = "<i>turn description:</i>\n%s" % currentStep['descriptionHtml']
+        message = "<i>turn description:</i>\n%s" % currentStep.getMessage()
         if self.dmod.hasNotificationSupport():
           self.dmod.notify(message,7000)
       #TODO: add support for modRana notifications once they support line wrapping
 
+  def _rerouteAuto(self):
+    """this function is called when automatic rerouting is triggered"""
+
+    # check time from last automatic reroute
+    dt = time.time() - self.lastAutomaticRerouteTimestamp
+    if dt >= AUTOMATIC_REROUTE_COUNTER_EXPIRATION_TIME:
+      # reset the automatic reroute counter
+      self.automaticRerouteCounter = 0
+      print('tbt: automatic reroute counter expired, clearing')
+
+    # on some routes, when we are moving away from the start of the route, it
+    # is needed to reroute a couple of times before the correct way is found
+    # on the other hand there should be a limit on the number of times
+    # modRana reroutes in a row
+    #
+    # SOLUTION:
+    # 1. enable automatic rerouting even though the route was not yet reached
+    # (as we are moving away from it)
+    # 2. do this only a limited number of times (up to 3 times in a row)
+    # 3. the counter is reset by manual rerouting, by reaching the route or after 10 minutes
+
+    if self.automaticRerouteCounter < MAX_CONSECUTIVE_AUTOMATIC_REROUTES:
+      print('tbt: faking that route was reached to enable new rerouting')
+      self.overrideRouteReached = True
+    else:
+      print('tbt: too many consecutive reroutes (%d),' % self.automaticRerouteCounter)
+      print('reach the route to enable automatic rerouting')
+      print('or reroute manually')
+
+    # increment the automatic reroute counter & update the timestamp
+    self.automaticRerouteCounter+=1
+    self.lastAutomaticRerouteTimestamp = time.time()
+
+    # trigger rerouting
+    self._reroute()
+
+  def _reroute(self):
+    # 1. say rerouting is in progress
+    voiceMessage = "rerouting"
+    voice = self.m.get('voice', None)
+    if voice:
+      voice.say(voiceMessage, "en") # make sure rerouting said with english voice
+    time.sleep(2) #TODO: improve this
+    # 2. get a new route from current position to destination
+    self.sendMessage("ms:route:reroute:fromPosToDest")
+    # 3. restart routing for to this new route from the closest point
+    self.sendMessage("ms:turnByTurn:start:closest")
+
   def drawMapOverlay(self,cr):
-      if self.steps:
+      if self.route:
         # get current step
         currentStep = self.getCurrentStep()
         proj = self.m.get('projection', None)
         # draw the current step indicator circle
         if currentStep and proj:
-          lat = currentStep['Point']['coordinates'][1]
-          lon = currentStep['Point']['coordinates'][0]
+          (lat, lon) = currentStep.getLL()
           (pointX, pointY) = proj.ll2xy(lat, lon)
           cr.set_source_rgb(1, 0, 0)
           cr.set_line_width(4)
@@ -116,7 +182,7 @@ class turnByTurn(ranaModule):
           cr.fill()
 
   def drawScreenOverlay(self, cr):
-    if self.steps: # is there something relevant to draw ?
+    if self.route: # is there something relevant to draw ?
     
       # get current step
       currentStep = self.getCurrentStep()
@@ -133,7 +199,6 @@ class turnByTurn(ranaModule):
 
         # construct parametric background for the cairo drawn buttons
         background="generic:;0;;1;5;0"
-#        background="generic:;;;;;"
         
         if self.navigationBoxHidden:
           # * show button
@@ -153,13 +218,16 @@ class turnByTurn(ranaModule):
           layout = pg.create_layout()
 
           # get the current turn message
-          message = currentStep['descriptionHtml']
+          message = currentStep.getMessage()
       
           # display current distance to the next point & other unit conversions
           units = self.m.get('units', None)
           if units and self.currentDistance:
             distString = units.m2CurrentUnitString(self.currentDistance,1,True)
-            currentDistString = units.m2CurrentUnitString(currentStep['mDistanceFromStart'],1,True)
+            if currentStep.getDistanceFromStart():
+              currentDistString = units.m2CurrentUnitString(currentStep.getDistanceFromStart(),1,True)
+            else:
+              currentDistString = "?"
             routeLengthString = units.m2CurrentUnitString(self.mRouteLength,1,True)
           else:
             distString = ""
@@ -179,7 +247,7 @@ class turnByTurn(ranaModule):
           layout.set_font_description(pango.FontDescription("Sans Serif 24")) #TODO: custom font size ?
           (lw,lh) = layout.get_size()
           if lw == 0 or lh == 0:
-            # no need to draw a zero are layout
+            # no need to draw a zero area layout
             return
 
           # get coordinates for the area available for text
@@ -202,7 +270,7 @@ class turnByTurn(ranaModule):
 
             if cut:
               """ notify the user that a part of the text was cut,
-              by drawing a red line and a scissors icon"""
+                          by drawing a red line and a scissors icon"""
               # draw the red line
               cr.set_source_rgb(1,0,0)
               cr.set_line_width(bh*0.01)
@@ -215,7 +283,6 @@ class turnByTurn(ranaModule):
               #TODO: show the whole message in a notifications after clicking the scissors
               # (this needs line wrapping support in modRana notifications)
 
-
             # clip out the overflowing part of the text
             cr.rectangle(ulX,ulY,usableWidth,clipHeight)
             cr.translate(ulX,ulY)
@@ -227,7 +294,11 @@ class turnByTurn(ranaModule):
 
           # use the bottom of the infobox to display info
           (bottomX,bottomY) = (bx, by+bh-6*border)
-          note = "%s/%s, %d/%d   <sub> tap this box to reroute</sub>" % (currentDistString, routeLengthString, self.currentStepIndex+1,len(self.steps))
+          if self.routeReached and self._automaticReroutingEnabled():
+            arString = "automatic rerouting enabled"
+          else:
+            arString ="tap this box to reroute"
+          note = "%s/%s, %d/%d   <sub> %s</sub>" % (currentDistString, routeLengthString, self.currentStepIndex+1, self.getMaxStepIndex()+1, arString)
           menus.drawText(cr, "%s" % note, bottomX, bottomY, bw, 6*border, 0, rgbaColor=self.navigationBoxText)
           # make clickable
           clickHandler = self.m.get('clickHandler', None)
@@ -246,9 +317,11 @@ class turnByTurn(ranaModule):
           # * hide button
           menus.drawButton(cr, bx+2*switchButtonWidth, by, hideButtonWidth, buttonStripOffset, "", "center:hide;0.1>%s" % background, "turnByTurn:toggleBoxHiding")
 
+  def _automaticReroutingEnabled(self):
+    return self.get('reroutingThreshold', REROUTING_DEFAULT_THRESHOLD)
 
   def sayTurn(self,message,distanceInMeters,forceLanguageCode=False):
-    """say a text-to-speach message about a turn
+    """say a text-to-speech message about a turn
        this basically wraps the simple say method from voice and adds some more information,
        like current distance to the turn
        """
@@ -265,12 +338,13 @@ class turnByTurn(ranaModule):
         # TODO: language specific distance strings
       text = distString + message
 
-      """ the message can contain unicode, this might cause an exception when printing it
-      in some systems (SHR-u on Neo, for example)"""
-      try:
-        print "saying: %s" % text
-      except UnicodeEncodeError:
-        print "voice: printing the current message to stdout failed do to unicode conversion error"
+#      """ the message can contain unicode, this might cause an exception when printing it
+#      in some systems (SHR-u on Neo, for example)"""
+#      try:
+#        print("saying: %s" % text)
+#        pass
+#      except UnicodeEncodeError:
+#        print("voice: printing the current message to stdout failed do to unicode conversion error")
 
       if forceLanguageCode:
         espeakLanguageCode = forceLanguageCode
@@ -279,79 +353,100 @@ class turnByTurn(ranaModule):
         espeakLanguageCode = self.get('directionsLanguage', 'en en').split(" ")[0]
       return voice.say(text,espeakLanguageCode)
 
+  def getMaxStepIndex(self):
+    return self.route.getMessagePointCount() - 1
+
   def getStartingStep(self, which='first'):
-    if self.steps:
+    if self.route:
       if which == 'first':
-        return self.steps[0]
+        return self.getStep(0)
       if which == 'closest':
         return self.getClosestStep()
 
   def getClosestStep(self):
+    """get the geographically closest step"""
     proj = self.m.get('projection', None) # we also need the projection module
     pos = self.get('pos', None) # and current position
     if pos and proj:
       (lat1,lon1) = pos
-      tempSteps = self.steps
+      tempSteps = self.route.getMessagePoints()
       for step in tempSteps:
-        lat2 = step['Point']['coordinates'][1]
-        lon2 = step['Point']['coordinates'][0]
-        step['currentDistance'] = geo.distance(lat1,lon1,lat2,lon2)*1000 # km to m
-      closestStep = sorted(tempSteps, key=lambda x: x['currentDistance'])[0]
+        (lat2, lon2) = step.getLL()
+        step.setCurrentDistance = geo.distance(lat1,lon1,lat2,lon2)*1000 # km to m
+      closestStep = sorted(tempSteps, key=lambda x: x.getCurrentDistance())[0]
+
       return closestStep
 
-  def getStep(self, id):
+  def getStep(self, index):
     """return steps for valid index, None otherwise"""
-    maxIndex = len(self.steps) - 1
-    if id > maxIndex or id < -(maxIndex+1):
-      print "wrong turn index: %d, max index is: %d" % (id, maxIndex)
+    maxIndex = self.getMaxStepIndex()
+    if index > maxIndex or index < -(maxIndex+1):
+      print("turnByTurn: wrong turn index: %d, max index is: %d" % (index, maxIndex))
       return None
     else:
-      return self.steps[id]
+      return self.route.getMessagePointByID(index)
 
   def setStepAsCurrent(self, step):
     """set a given step as current step"""
-    id = step['id']
-    self.currentStepIndex = id
+    id = self.route.getMessagePointID(step)
+    self._setCurrentStepIndex(id)
 
   def getCurrentStep(self):
     """return current step"""
-    return self.steps[self.currentStepIndex]
+    return self.route.getMessagePointByID(self.currentStepIndex)
+
+  def getStepID(self, step):
+    return self.route.getMessagePointID(step)
 
   def getCurrentStepVisitStatus(self):
     """report visit status for  current step"""
-    return self.steps[self.currentStepIndex]['visited']
+    return self.getCurrentStep().getVisited()
 
   def markCurrentStepAsVisited(self):
     """mark current step as visited"""
-    self.steps[self.currentStepIndex]['visited'] = True
+    self.getCurrentStep().setVisited(True)
 
   def switchToPreviousStep(self):
     """switch to previous step and clean up"""
     nextIndex = self.currentStepIndex - 1
     if nextIndex >= 0:
-      self.currentStepIndex = nextIndex
+      self._setCurrentStepIndex(nextIndex)
       self.espeakFirstTrigger = False
       self.espeakSecondTrigger = False
-      print "switching to previous step"
+      print("tbt: switching to previous step")
     else:
-      print "previous step reached"
+      print("tbt: previous step reached")
 
   def switchToNextStep(self):
     """switch to next step and clean up"""
-    maxIndex = len(self.steps) - 1
+    maxIndex = self.getMaxStepIndex()
     nextIndex = self.currentStepIndex + 1
     if nextIndex <= maxIndex:
-      self.currentStepIndex = nextIndex
+      self._setCurrentStepIndex(nextIndex)
       self.espeakFirstAndHalfTrigger = False
       self.espeakFirstTrigger = False
       self.espeakSecondTrigger = False
-      print "switching to next step"
+      print("tbt: switching to next step")
     else:
-      print "last step reached"
+      print("tbt: last step reached")
+      self._lastStepReached()
+
+  def _lastStepReached(self):
+    """handle all tasks that are needed once the last step is reached"""
+    #disable automatic rerouting
+    self._stopTBTWorker()
+    # automatic rerouting needs to be disabled to prevent rerouting
+    # once the destination was reached
+
+
+
+  def _setCurrentStepIndex(self, index):
+    self.currentStepIndex = index
+    self._doNavigationUpdate()
 
   def enabled(self):
     """return True if enabled, false otherwise"""
-    if self.steps:
+    if self.route:
       return True
     else:
       return False
@@ -365,32 +460,28 @@ class turnByTurn(ranaModule):
     """NOTE: turn and step are used interchangeably in the documentation"""
     m = self.m.get('route', None)
     if m:
-      (dirs,routeRequestSentTimestamp) = m.getCurrentDirections()
-      if dirs: # is the route nonempty ?
+      (route,routeRequestSentTimestamp) = m.getCurrentDirections()
+      if route: # is the route nonempty ?
+        self.route = route
+        # get route in radians for automatic rerouting
+        self.radiansRoute = route.getPointsLLERadians(dropElevation=True)
+        # start rerouting watch
+        self._startTBTWorker()
+
         # show the warning message
         self.sendMessage('ml:notification:m:use at own risk, watch for cliffs, etc.;3')
-        route = dirs['Directions']['Routes'][0]
         # for some reason the combined distance does not account for the last step
-        self.mRouteLength = route['Distance']['meters'] + route['Steps'][-1]["Distance"]["meters"]
-        self.steps = []
-        mDistanceFromStart = route['Steps'][-1]["Distance"]["meters"]
-        for step in route['Steps']:
-          step['currentDistance'] = None # add the currentDistance key
-          # add and compute the distance from start
-          step['mDistanceFromStart'] = mDistanceFromStart
-          mDistanceFromLast = step["Distance"]["meters"]
-          mDistanceFromStart = mDistanceFromStart + mDistanceFromLast
-          self.steps.append(step)
-        self.steps = dirs['Directions']['Routes'][0]['Steps']
+        self.mRouteLength = route.getLength()
+
         # some statistics
         metersPerSecSpeed = self.get('metersPerSecSpeed', None)
         dt = time.time() - routeRequestSentTimestamp
-        print "route lookup took: %f s" % dt
+        print("turnByTurn: route lookup took: %f s" % dt)
         if dt and metersPerSecSpeed:
           dm = dt * metersPerSecSpeed
-          print "distance traveled during lookup: %f m" % dm
+          print("distance traveled during lookup: %f m" % dm)
         """the duration of the road lookup and other variables are currently not used
-        in the heuristics but might be added later to make the heursitics more robust"""
+        in the heuristics but might be added later to make the heuristics more robust"""
 
 
         """now we decide if we use the closest turn, or the next one,
@@ -398,63 +489,61 @@ class turnByTurn(ranaModule):
         cs = self.getClosestStep() # get geographically closest step
         pos = self.get('pos', None) # get current position
         pReachedDist = int(self.get('pointReachedDistance', 30)) # get the trigger distance
-        nextTurnId = cs['id'] + 1
+        nextTurnId = self.getStepID(cs) + 1
         nextStep = self.getStep(nextTurnId)
         # check if we have all the data needed for our heuristics
-        print "tbt: trying to guess correct step to start navigation"
+        print("tbt: trying to guess correct step to start navigation")
         if nextStep and pos and pReachedDist:
           (lat,lon) = pos
-          (csLat,csLon) = (cs['Point']['coordinates'][1],cs['Point']['coordinates'][0])
-          (nsLat,nsLon) = (nextStep['Point']['coordinates'][1],nextStep['Point']['coordinates'][0])
+          (csLat,csLon) = cs.getLL()
+          (nsLat,nsLon) = nextStep.getLL()
           pos2nextStep = geo.distance(lat,lon,nsLat,nsLon)*1000
           pos2currentStep = geo.distance(lat,lon,csLat,csLon)*1000
           currentStep2nextStep = geo.distance(csLat,csLon,nsLat,nsLon)*1000
 #          print "pos",(lat,lon)
 #          print "cs",(csLat,csLon)
 #          print "ns",(nsLat,nsLon)
-          print "position to next turn: %f m" % pos2nextStep
-          print "position to current turn: %f m" % pos2currentStep
-          print "current turn to next turn: %f m" % currentStep2nextStep
-          print "turn reached trigger distance: %f m" % pReachedDist
+          print("position to next turn: %f m" % pos2nextStep)
+          print("position to current turn: %f m" % pos2currentStep)
+          print("current turn to next turn: %f m" % currentStep2nextStep)
+          print("turn reached trigger distance: %f m" % pReachedDist)
 
           if pos2currentStep > pReachedDist:
-            """this means we are out of the "capture circle" of the closest step"""
+            #this means we are out of the "capture circle" of the closest step
 
-            """what is more distant, the closest or the next step ?"""
+            # what is more distant, the closest or the next step ?
             if pos2nextStep < currentStep2nextStep:
-              """we are mostly probably already past the closest step,
-              so we switch to the next step at once"""
-              print "tbt: already past closest turn, switching to next turn"
+              # we are mostly probably already past the closest step,
+              # so we switch to the next step at once
+              print("tbt: already past closest turn, switching to next turn")
               self.setStepAsCurrent(nextStep)
               """we play the message for the next step,
               with current distance to this step,
               to assure there is some voice output immediately after
               getting a new route or rerouting"""
-              plaintextMessage = nextStep['descriptionEspeak']
+              plaintextMessage = nextStep.getSSMLMessage()
               self.sayTurn(plaintextMessage, pos2nextStep)
-
-
             else:
-              """we have probably not yet reached the closest step,
-                 so we start navigation from it"""
-              print "tbt: closest turn not yet reached"
+              # we have probably not yet reached the closest step,
+              # so we start navigation from it
+              print("tbt: closest turn not yet reached")
               self.setStepAsCurrent(cs)
 
           else:
             """we are inside the  "capture circle" of the closest step,
-            this meens the navigation will trigger the voice message by itself
+            this means the navigation will trigger the voice message by itself
             and correctly switch to next step
             -> no need to switch to next step from here"""
-            print "tbt: inside reach distance of closest turn"
+            print("tbt: inside reach distance of closest turn")
             self.setStepAsCurrent(cs)
 
         else:
           """we dont have some of the data, that is needed to decide
           if we start the navigation from the closest step of from the step that is after it
           -> we just start from the closest step"""
-          """tbt: not enough data to decide, using closest turn"""
+          print("tbt: not enough data to decide, using closest turn")
           self.setStepAsCurrent(cs)
-    self.doNavigationUpdate() # run a first time navigation update
+    self._doNavigationUpdate() # run a first time navigation update
     self.locationWatchID = self.watch('locationUpdated', self.locationUpdateCB)
     print("tbt: started")
       
@@ -463,33 +552,33 @@ class turnByTurn(ranaModule):
     # remove location watch
     if self.locationWatchID:
       self.removeWatch(self.locationWatchID)
-    
+    # cleanup
     self.goToInitialState()
+    self._stopTBTWorker()
     print("tbt: stopped")
 
   def locationUpdateCB(self, key, newValue, oldValue):
     """position changed, do a tbt navigation update"""
     if key == "locationUpdated": # just to be sure
-      self.doNavigationUpdate()
+      self._doNavigationUpdate()
     else:
-      print "tbt: invalid key: %r" % key
+      print("tbt: invalid key: %r" % key)
 
-  def doNavigationUpdate(self):
+  def _doNavigationUpdate(self):
     """do a navigation update"""
     # make sure there really are some steps
-    if not self.steps:
-      print "tbt: error no navigation steps"
+    if not self.route:
+      print("tbt: error no route")
       return
     pos = self.get('pos', None)
-    if pos == None:
-      print "tbt: skipping update, invalid position"
+    if pos is None:
+      print("tbt: skipping update, invalid position")
       return
 
     # get/compute/update necessary the values
     (lat1,lon1) = pos
     currentStep = self.getCurrentStep()
-    lat2 = currentStep['Point']['coordinates'][1]
-    lon2 = currentStep['Point']['coordinates'][0]
+    lat2, lon2 = currentStep.getLL()
     currentDistance = geo.distance(lat1,lon1,lat2,lon2)*1000 # km to m
     self.currentDistance = currentDistance # update current distance
 
@@ -540,9 +629,16 @@ class turnByTurn(ranaModule):
 
     if metersPerSecSpeed:
       # check if we can miss the point by going too fast -> mps speed > point reached distance
+      # also enlarge the rerouting threshold as it looks like it needs to be larger
+      # when moving at high speed to prevent unnecessary rerouting
       if metersPerSecSpeed > pointReachedDistance*0.75:
         pointReachedDistance = metersPerSecSpeed*2
-        print "tbt: enlarging point reached distance to: %1.2f m due to large speed (%1.2f m/s)" % (pointReachedDistance, metersPerSecSpeed)
+#        print("tbt: enlarging point reached distance to: %1.2f m due to large speed (%1.2f m/s)" % (pointReachedDistance, metersPerSecSpeed))
+
+      if metersPerSecSpeed > INCREASE_REROUTING_THRESHOLD_SPEED:
+        self.reroutingThresholdMultiplier = REROUTING_THRESHOLD_MULTIPLIER
+      else:
+        self.reroutingThresholdMultiplier = 1.0
 
       # speed & time based triggering
       lowSpeed = float(self.get('minAnnounceSpeed', 13.89))
@@ -559,22 +655,22 @@ class turnByTurn(ranaModule):
       distance = max(distance, warnTime * metersPerSecSpeed)
 
       if self.get('debugTbT', False):
-        print "#####"
-        print "min/max announce time: %d/%d s" % (lowTime, highTime)
-        print "trigger distance: %1.2f m (%1.2f s warning)" % (distance, distance/float(metersPerSecSpeed))
-        print "current distance: %1.2f m" % currentDistance
-        print "current speed: %1.2f m/s (%1.2f km/h)" % (metersPerSecSpeed, metersPerSecSpeed*3.6)
-        print "point reached distance: %f m" % pointReachedDistance
-        print "1. triggered=%r, 1.5. triggered=%r, 2. triggered=%r" % (self.espeakFirstTrigger, self.espeakFirstAndHalfTrigger, self.espeakSecondTrigger)
+        print("#####")
+        print("min/max announce time: %d/%d s" % (lowTime, highTime))
+        print("trigger distance: %1.2f m (%1.2f s warning)" % (distance, distance/float(metersPerSecSpeed)))
+        print("current distance: %1.2f m" % currentDistance)
+        print("current speed: %1.2f m/s (%1.2f km/h)" % (metersPerSecSpeed, metersPerSecSpeed*3.6))
+        print("point reached distance: %f m" % pointReachedDistance)
+        print("1. triggered=%r, 1.5. triggered=%r, 2. triggered=%r" % (self.espeakFirstTrigger, self.espeakFirstAndHalfTrigger, self.espeakSecondTrigger))
         if warnTime > 30:
-          print "optional (20 s) trigger distance: %1.2f" % (20.0*metersPerSecSpeed)
+          print("optional (20 s) trigger distance: %1.2f" % (20.0*metersPerSecSpeed))
 
       if currentDistance <= pointReachedDistance:
         """this means we reached the point"""
         if self.espeakSecondTrigger == False:
-          print "triggering espeak nr. 2"
+          print("triggering espeak nr. 2")
           # say the message without distance
-          plaintextMessage = currentStep['descriptionEspeak']
+          plaintextMessage = currentStep.getSSMLMessage()
           # consider turn said even if it was skipped (ignore errors)
           self.sayTurn(plaintextMessage, 0)
           self.markCurrentStepAsVisited() # mark this point as visited
@@ -585,8 +681,8 @@ class turnByTurn(ranaModule):
         if currentDistance <= distance:
           """this means we reached an optimal distance for saying the message"""
           if self.espeakFirstTrigger == False:
-            print "triggering espeak nr. 1"
-            plaintextMessage = currentStep['descriptionEspeak']
+            print("triggering espeak nr. 1")
+            plaintextMessage = currentStep.getSSMLMessage()
             if self.sayTurn(plaintextMessage, currentDistance):
               self.espeakFirstTrigger = True # first message done
         if self.espeakFirstAndHalfTrigger == False and warnTime > 30:
@@ -594,11 +690,119 @@ class turnByTurn(ranaModule):
             """in case that the warning time gets too big, add an intermediate warning at 20 seconds
             NOTE: this means it is said after the first trigger
             """
-            plaintextMessage = currentStep['descriptionEspeak']
+            plaintextMessage = currentStep.getSSMLMessage()
             if self.sayTurn(plaintextMessage, currentDistance):
               self.espeakFirstAndHalfTrigger = True # intermediate message done
 
-              
+      ## automatic rerouting ##
+
+      # is automatic rerouting enabled from options
+      # enabled == threshold that is not not None
+      if self._automaticReroutingEnabled():
+        # rerouting is enabled only once the route is reached for the first time
+        if self.onRoute and not self.routeReached:
+          self.routeReached = True
+          self.automaticRerouteCounter = 0
+          print('tbt: route reached, rerouting enabled')
+
+        # did the TBT worker detect that the rerouting threshold was reached ?
+        if self._reroutingConditionsMet():
+          # test if enough consecutive divergence point were recorded
+          if self.reroutingThresholdCrossedCounter >= REROUTING_TRIGGER_COUNT:
+            # reset the routeReached override
+            self.overrideRouteReached = False
+            # trigger rerouting
+            self._rerouteAuto()
+        else:
+          # reset the counter
+          self.reroutingThresholdCrossedCounter = 0
+
+  def _reroutingConditionsMet(self):
+    return (self.routeReached or self.overrideRouteReached) and not self.onRoute
+
+  def _followingRoute(self):
+    """are we still following the route or is rerouting needed"""
+    start1 = time.clock()
+    pos = self.get('pos', None)
+    proj = self.m.get('projection', None)
+    if pos and proj:
+      pLat, pLon = pos
+      # we use Radians to get rid of radian conversion overhead for
+      # the geographic distance computation method
+      radiansLL = self.radiansRoute
+      pLat = geo.radians(pLat)
+      pLon = geo.radians(pLon)
+      if len(radiansLL) == 0:
+        print("Divergence: can't follow a zero point route")
+        return False
+      elif len(radiansLL) == 1: # 1 point route
+        aLat, aLon = radiansLL[0]
+        minDistance = geo.distanceApproxRadians(pLat, pLon, aLat, aLon)
+      else: # 2+ points route
+        aLat, aLon = radiansLL[0]
+        bLat, bLon = radiansLL[1]
+        minDistance = geo.distancePointToLineRadians(pLat, pLon, aLat, aLon, bLat, bLon)
+        aLat, aLon = bLat, bLon
+        for point in radiansLL[1:]:
+          bLat, bLon = point
+          dist = geo.distancePointToLineRadians(pLat, pLon, aLat, aLon, bLat, bLon)
+          if dist < minDistance:
+            minDistance = dist
+          aLat, aLon = bLat, bLon
+      # the multiplier tries to compensate for high speed movement
+      threshold = float(self.get('reroutingThreshold', REROUTING_DEFAULT_THRESHOLD))*self.reroutingThresholdMultiplier
+      print("Divergence from route: %1.2f/%1.2f m computed in %1.0f ms" % (minDistance*1000, float(threshold), (1000 * (time.clock() - start1))) )
+      return minDistance*1000 < threshold
+
+  def _startTBTWorker(self):
+    print "tbt: starting worker thread"
+    startThread = True
+    if not self.TBTWorker: # reuse previous thread or start new one
+      self.TBTWorkerEnabled = True
+      t = Thread(target=self._TBTWorker)
+      t.daemon = True
+      t.start()
+      self.TBTWorker = t
+    else:
+      print "tbt: reusing worker thread"
+
+  def _stopTBTWorker(self):
+    self.TBTWorkerEnabled = False
+    self.TBTWorker = None
+
+  def _TBTWorker(self):
+    """this function is run in its own thread and check if
+    we are following the current route"""
+    print("TBTWorker: started")
+    while self.route and self.TBTWorkerEnabled:
+      # first make sure automatic rerouting is enabled
+      # eq. reroutingThreshold != None
+      if self._automaticReroutingEnabled():
+        # check if we are still following the route
+#        print('TBTWorker: checking divergence from route')
+        self.onRoute = self._followingRoute()
+        if self._reroutingConditionsMet():
+          print('TBTWorker: divergence detected')
+          # switch to quick updates
+          for i in range(0, REROUTING_TRIGGER_COUNT+1):
+            time.sleep(1)
+            onRoute = self._followingRoute()
+            if onRoute: # divergence stopped
+              self.onRoute = onRoute
+              print('TBTWorker: false alarm')
+              break
+            else: # still diverging from current route
+              self.onRoute = onRoute
+              # increase divergence counter
+              self.reroutingThresholdCrossedCounter+=1
+              print('TBTWorker: increasing divergence counter (%d)' % self.reroutingThresholdCrossedCounter)
+      time.sleep(REROUTE_CHECK_INTERVAL/1000.0)
+    print("TBTWorker: shutting down")
+
+  def shutdown(self):
+    # cleanup
+    self.stopTBT()
+
 if(__name__ == "__main__"):
   a = example({}, {})
   a.update()
